// ==================================================================================
// Chuck step sequencer v2 (for Breakage v2)
// MIDI out version
// ==================================================================================

// ==================================================================================
// osc 
// ==================================================================================

3000 => int oscSendPort;
2000 => int oscRecvPort;

// create OSC sender
OscSend oscXmit;
oscXmit.setHost( "localhost", oscSendPort);

// create OSC receiver
OscRecv oscRecv;
oscRecvPort => oscRecv.port;
oscRecv.listen();
oscRecv.event( "/setStepVelocity,i,i,f" ) @=> OscEvent oscStepVelocityEvent;
oscRecv.event( "/setStepPitch,i,i,i" ) @=> OscEvent oscStepPitchEvent;
oscRecv.event( "/setBpm,f" ) @=> OscEvent oscBpmEvent;
oscRecv.event( "/setChannelVolume,i,f" ) @=> OscEvent oscChannelVolumeEvent;
oscRecv.event( "/setChannelSample,i,s" ) @=> OscEvent oscChannelSampleEvent;
oscRecv.event( "/setSwing,i,s" ) @=> OscEvent oscSwingEvent;
oscRecv.event( "/setPatternDimensions,i,i" ) @=> OscEvent oscPatternDimensionEvent;
oscRecv.event( "/play,i" ) @=> OscEvent oscPlayEvent;

// ==================================================================================
// midi 
// ==================================================================================

// data structures
MidiOut mout;
MidiMsg msg;

// connect
if( !mout.open( 0 ) ) me.exit();
<<< "connected to MIDI port 0", "" >>>;

// ==================================================================================
// timer 
// ==================================================================================

// sequencer's step, cycle and state
0 => int step;
0 => int cycle;
0 => int playState;		// 0 = play, 1 = stop, 2 = pause

// prepare timer 
120 => float bpm;
60 / bpm => float bps;
bps::second => dur timer;
timer - (now % timer) => now;

// prepare swing
0.2 => float swingTo;
-0.2 => float swingFro;
bps => float thisBps;

1 => int swing;

<<< "<breakage channels=\"", channels, "\" length=\"", channelLength, "\" />" >>>;

<<<"<set bpm=", bpm, " />" >>>;

// ==================================================================================
// sequence volumes array
// ==================================================================================

4 => int channels;
16 => int channelLength;

float stepVolumes[channels][channelLength];
int stepPitches[channels][channelLength];

initDimensions(4, 16);

fun void initDimensions(int ch, int l) {

	// sequence size
	ch => channels;
	l => channelLength;
		
	// set default step pitches
	for( 0 => int i; i < stepPitches.cap(); i++ ) {
		for( 0 => int j; j < stepPitches[0].cap(); j++ ) {
			60 => stepPitches[i][j];
		}
	}
	
	// restart if necessary
	if(step > channelLength) 0 => step;
}

// ==================================================================================
// samples: sound buffers
// ==================================================================================

// prepare sound buffer and channel data
sndbuf soundBuffers[channels];
gain masterGain;

float channelVolumes[channels];
int channelRootNotes[channels];

// set channel volumes and root notes
for( 0 => int i; i < soundBuffers.cap(); i++ )
{
    soundBuffers[i] => masterGain;
    0.5 => channelVolumes[i];
	60 => channelRootNotes[i];
}

// load default samples
"samples/kick.wav" => soundBuffers[0].read;
"samples/snare.wav" =>  soundBuffers[1].read;
"samples/closedhat.wav" =>  soundBuffers[2].read;
"samples/openhat.wav" =>  soundBuffers[3].read;

// apply filter to channel 2 (closed hat)
soundBuffers[2] => biquad f => masterGain;

// global fx
PRCRev reverb;
Chorus chorus;

//masterGain => chorus => reverb => dac;
masterGain => dac;

0.8 => masterGain.gain;

0.15 => reverb.mix;

0.15 => chorus.modFreq;
0.15 => chorus.modDepth;
0.45 => chorus.mix;

// ==================================================================================
// samples: filter
// ==================================================================================

// set biquad pole radius
.99 => f.prad;
// set biquad gain
.05 => f.gain;
// set equal zeros 
1 => f.eqzs;
// our float
0.0 => float t;

// ==================================================================================
// sample: play pitch
// ==================================================================================

// set given sample buffer to play at given pitch
fun void setSampleRate(int buffer, int note) {
	channelRootNotes[buffer] => int root;
	std.mtof(note) / std.mtof(root) => soundBuffers[buffer].rate;
}

// ==================================================================================
// inst: sitar
// ==================================================================================

// instrument - NOT USED
Sitar sit => PRCRev r => dac;
.15 => r.mix;

// play function
fun void playSitar(float f) {
    f => sit.freq;

    // pluck!
    std.rand2f( 0.4, 0.9 ) => sit.noteOn;
}

// ==================================================================================
// osc listener loops
// ==================================================================================

// step pitch
fun void stepPitchListener() {
	while ( true ) {
	    oscStepPitchEvent => now;  // wait for event to arrive
	
	    while ( oscStepPitchEvent.nextMsg() != 0 ) { 
	        oscStepPitchEvent.getInt() => int a;
	        oscStepPitchEvent.getInt() => int b;
	        oscStepPitchEvent.getInt() => int c;
	        c => stepPitches[a][b];
	    }
	}
}
spork ~ stepPitchListener();

// step volume
fun void stepVelocityListener() {
	while ( true ) {
	    oscStepVelocityEvent => now;  // wait for event to arrive
	
	    while ( oscStepVelocityEvent.nextMsg() != 0 ) { 
	        oscStepVelocityEvent.getInt() => int a;
	        oscStepVelocityEvent.getInt() => int b;
	        oscStepVelocityEvent.getFloat() => float c;
	        c => stepVolumes[a][b];
	    }
	}
}
spork ~ stepVelocityListener();

// channel volume
fun void channelVolumeListener() {
	while ( true ) {
	    oscChannelVolumeEvent => now;  // wait for event to arrive
	
	    while ( oscChannelVolumeEvent.nextMsg() != 0 ) { 
	        oscChannelVolumeEvent.getInt() => int a;
	        oscChannelVolumeEvent.getFloat() => float b;
	        b => channelVolumes[a];
	        <<<"<set channel=\"", a, "\" volume=\"",b,"\" />" >>>;
	    }
	}
}
spork ~ channelVolumeListener();

// bpm
fun void bpmListener() {
	while ( true ) {
	    oscBpmEvent => now; // wait for event to arrive
	
	    while ( oscBpmEvent.nextMsg() != 0 ) { 
	        oscBpmEvent.getFloat() => bpm;
			60 / bpm => bps;
			bps::second => dur timer;
			timer - (now % timer) => now;
			<<<"<set bpm=", bpm, " />" >>>;
	    }
	}
}
spork ~ bpmListener();

// open sample
fun void setChannelSampleListener() {
	while ( true ) {
	    oscChannelSampleEvent => now;
	
	    // grab the next message from the queue. 
	    while ( oscChannelSampleEvent.nextMsg() != 0 ) { 
	        oscChannelSampleEvent.getInt() => int i;
	        oscChannelSampleEvent.getString() => string s;
	        s => soundBuffers[i].read;
	        <<<"<set channel=\"", i, "\" sample=\"",s,"\" />" >>>;
	    }
	}
}
spork ~ setChannelSampleListener();

// play state
fun void startPlayListener() {
	while ( true ) {
	    oscPlayEvent => now;
	
	    // grab the next message from the queue. 
	    while ( oscPlayEvent.nextMsg() != 0 ) { 
			oscPlayEvent.getInt() => int i;
			
			// 0 = play, 1 = stop, 2 = pause
			
			if(i == 1) {
				// if we're stopping, reset the loop position
				1 => playState;
				0 => step;
				
			} else if(i == 2) {
				// if we're pausing, check state
				if(playState == 0) {
					2 => playState;
				} else {
					0 => playState;
				}
			
			} else {
				// playing, so pause or stop
				i => playState;
			}
	    }
	}
}
spork ~ startPlayListener();

// pattern dimension
fun void startPatternDimensionListener() {
	while ( true ) {
	    oscPatternDimensionEvent => now;
	
	    // grab the next message from the queue. 
	    while ( oscPatternDimensionEvent.nextMsg() != 0 ) { 
			oscPatternDimensionEvent.getInt() => int i;
			oscPatternDimensionEvent.getInt() => int j;
			
			initDimensions(i, j);
	    }
	}
}
spork ~ startPatternDimensionListener();

// =============================
// filter loop
// =============================

fun void filter() {
	//while( true ) {
    	// sweep the filter resonant frequency
	    //100.0 + std.abs(math.sin(t)) * 1000.0 => f.pfreq;
    	//t + .01 => t;
	    // advance time
    	//100::ms => now;
	//}
}
//	spork ~ filter();

// =============================
// step seq loop
// =============================

<<< "<breakage channels=\"", channels, "\" length=\"", channelLength, "\" />" >>>;
<<< "<set bpm=", bpm, " />" >>>;
<<< "<step>" >>>;

// time loop
while( true ) {

	// are we paused? 
	while(playState != 0) {
		0.05::second => now;
	}
	
	// print step & cycle
	<<< "</step>" >>>;
	<<< "<step step=\"", step ,"\" cycle=\"", cycle, "\" >" >>>;
	
	// play sound - send MIDI note on messages
	for( 0 => int i; i < soundBuffers.cap(); i++ )
	{
		//if(stepVolumes[i][step] != 0) {
			<<< "<play channel=\"", i,"\" volume=\"", stepVolumes[i][step], "\" pitch=\"" , stepPitches[i][step], "\" />" >>>;

			// create msg
			0x90 + i => msg.data1;						// note on
			stepPitches[i][step] => msg.data2;		// pitch
			(stepVolumes[i][step] * 127) $ int => int v;
			v => msg.data3;							// velocity

			// send
			mout.send( msg );

//			0 => soundBuffers[i].pos;
//			setSampleRate(i, stepPitches[i][step]);
//		    stepVolumes[i][step] * channelVolumes[i] => soundBuffers[i].gain;
		//}
	}
	
	// advance time
	
//	if(swing == 1) {
//		0 => swing;
//		bps + swingTo => thisBps;
//	} else {
//		1 => swing;
//		bps + swingFro => thisBps;
//	}
    
    bps::timer => now;
    
    // send step message
    oscXmit.startMsg( "/currentStep", "i" );
	step => oscXmit.addInt;
    
    // increase step
	1 +=> step;

	// loop sequence if we've reached the end
	if( step == channelLength ) {
		0 => step;
		1 +=> cycle;
	}
}